import os
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from numpy.lib._iotools import NameValidator
from pandas.testing import assert_series_equal
from pastas.stats.tests import runs_test, stoffer_toloi
from tqdm.auto import tqdm

from pastastore.version import PASTAS_LEQ_022


def _custom_warning(message, category=UserWarning, filename="", lineno=-1, *args):
    print(f"{filename}:{lineno}: {category.__name__}: {message}")


class ItemInLibraryException(Exception):
    pass


def delete_pystore_connector(
    conn=None,
    path: Optional[str] = None,
    name: Optional[str] = None,
    libraries: Optional[List[str]] = None,
) -> None:  # pragma: no cover
    """Delete libraries from pystore.

    Parameters
    ----------
    conn : PystoreConnector, optional
        PystoreConnector object
    path : str, optional
        path to pystore
    name : str, optional
        name of the pystore
    libraries : Optional[List[str]], optional
        list of library names to delete, by default None which deletes
        all libraries
    """
    import pystore

    if conn is not None:
        name = conn.name
        path = conn.path
    elif name is None or path is None:
        raise ValueError("Please provide 'name' and 'path' OR 'conn'!")

    print(f"Deleting PystoreConnector database: '{name}' ...", end="")
    pystore.set_path(path)
    if libraries is None:
        pystore.delete_store(name)
        print(" Done!")
    else:
        store = pystore.store(name)
        for lib in libraries:
            print()
            store.delete_collection(lib)
            print(f" - deleted: {lib}")


def delete_arctic_connector(
    conn=None,
    connstr: Optional[str] = None,
    name: Optional[str] = None,
    libraries: Optional[List[str]] = None,
) -> None:  # pragma: no cover
    """Delete libraries from arctic database.

    Parameters
    ----------
    conn : pastastore.ArcticConnector
        ArcticConnector object
    connstr : str, optional
        connection string to the database
    name : str, optional
        name of the database
    libraries : Optional[List[str]], optional
        list of library names to delete, by default None which deletes
        all libraries
    """
    import arctic

    if conn is not None:
        name = conn.name
        connstr = conn.connstr
    elif name is None or connstr is None:
        raise ValueError("Provide 'name' and 'connstr' OR 'conn'!")

    arc = arctic.Arctic(connstr)

    print(f"Deleting ArcticConnector database: '{name}' ... ", end="")
    # get library names
    if libraries is None:
        libs = []
        for ilib in arc.list_libraries():
            if ilib.split(".")[0] == name:
                libs.append(ilib)
    elif name is not None:
        libs = [name + "." + ilib for ilib in libraries]
    else:
        raise ValueError("Provide 'name' and 'connstr' OR 'conn'!")

    for lib in libs:
        arc.delete_library(lib)
        if libraries is not None:
            print()
            print(f" - deleted: {lib}")
    print("Done!")


def delete_arcticdb_connector(
    conn=None,
    uri: Optional[str] = None,
    name: Optional[str] = None,
    libraries: Optional[List[str]] = None,
) -> None:
    """Delete libraries from arcticDB database.

    Parameters
    ----------
    conn : pastastore.ArcticDBConnector
        ArcticDBConnector object
    uri : str, optional
        uri connection string to the database
    name : str, optional
        name of the database
    libraries : Optional[List[str]], optional
        list of library names to delete, by default None which deletes
        all libraries
    """
    import shutil

    import arcticdb

    if conn is not None:
        name = conn.name
        uri = conn.uri
    elif name is None or uri is None:
        raise ValueError("Provide 'name' and 'uri' OR 'conn'!")

    arc = arcticdb.Arctic(uri)

    print(f"Deleting ArcticDBConnector database: '{name}' ... ", end="")
    # get library names
    if libraries is None:
        libs = []
        for ilib in arc.list_libraries():
            if ilib.split(".")[0] == name:
                # TODO: remove replace when arcticdb is able to delete
                libs.append(ilib.replace(".", "/"))
    elif name is not None:
        # TODO: replace / with . when arcticdb is able to delete
        libs = [name + "/" + ilib for ilib in libraries]
    else:
        raise ValueError("Provide 'name' and 'uri' OR 'conn'!")

    for lib in libs:
        # arc.delete_library(lib)  # TODO: not working at the moment.
        shutil.rmtree(os.path.join(conn.uri.split("//")[-1], lib))

        if libraries is not None:
            print()
            print(f" - deleted: {lib}")

    remaining = [ilib for ilib in arc.list_libraries() if ilib.split(".") == name]
    if len(remaining) == 0:
        shutil.rmtree(os.path.join(conn.uri.split("//")[-1], name))

    print("Done!")


def delete_dict_connector(conn, libraries: Optional[List[str]] = None) -> None:
    print(f"Deleting DictConnector: '{conn.name}' ... ", end="")
    if libraries is None:
        del conn
        print(" Done!")
    else:
        for lib in libraries:
            print()
            delattr(conn, f"lib_{conn.libname[lib]}")
            print(f" - deleted: {lib}")
    print("Done!")


def delete_pas_connector(conn, libraries: Optional[List[str]] = None) -> None:
    import shutil

    print(f"Deleting PasConnector database: '{conn.name}' ... ", end="")
    if libraries is None:
        shutil.rmtree(conn.path)
        print(" Done!")
    else:
        for lib in libraries:
            print()
            shutil.rmtree(os.path.join(conn.path, lib))
            print(f" - deleted: {lib}")
        print("Done!")


def delete_pastastore(pstore, libraries: Optional[List[str]] = None) -> None:
    """Delete libraries from PastaStore.

    Note
    ----
    This deletes the original PastaStore object. To access
    data that has not been deleted, it is recommended to create a new
    PastaStore object with the same Connector settings. This also creates
    new empty libraries if they were deleted.

    Parameters
    ----------
    pstore : pastastore.PastaStore
        PastaStore object to delete (from)
    libraries : Optional[List[str]], optional
        list of library names to delete, by default None which deletes
        all libraries

    Raises
    ------
    TypeError
        when Connector type is not recognized
    """
    if pstore.conn.conn_type == "pystore":
        delete_pystore_connector(conn=pstore.conn, libraries=libraries)
    elif pstore.conn.conn_type == "dict":
        delete_dict_connector(pstore)
    elif pstore.conn.conn_type == "arctic":
        delete_arctic_connector(conn=pstore.conn, libraries=libraries)
    elif pstore.conn.conn_type == "arcticdb":
        delete_arcticdb_connector(conn=pstore.conn, libraries=libraries)
    elif pstore.conn.conn_type == "pas":
        delete_pas_connector(conn=pstore.conn, libraries=libraries)
    else:
        raise TypeError(
            "Unrecognized pastastore Connector type: " f"{pstore.conn.conn_type}"
        )


def validate_names(
    s: Optional[str] = None,
    d: Optional[dict] = None,
    replace_space: Optional[str] = "_",
    deletechars: Optional[str] = None,
    **kwargs,
) -> Union[str, Dict]:
    """Remove invalid characters from string or dictionary keys.

    Parameters
    ----------
    s : str, optional
        remove invalid characters from string
    d : dict, optional
        remove invalid characters from keys from dictionary
    replace_space : str, optional
        replace spaces by this character, by default "_"
    deletechars : str, optional
        a string combining invalid characters, by default None

    Returns
    -------
    str, dict
        string or dict with invalid characters removed
    """
    validator = NameValidator(
        replace_space=replace_space, deletechars=deletechars, **kwargs
    )
    if s is not None:
        new_str = validator(s)  # tuple
        if len(new_str) == 1:
            return new_str[0]
        else:
            return new_str
    elif d is not None:
        new_dict = {}
        for k, v in d.items():
            new_dict[validator(k)[0]] = v
        return new_dict
    else:
        raise ValueError("Provide one of 's' (string) or 'd' (dict)!")


def compare_models(ml1, ml2, stats=None, detailed_comparison=False):
    """Compare two Pastas models.

    Parameters
    ----------
    ml1 : pastas.Model
        first model to compare
    ml2 : pastas.Model
        second model to compare
    stats : list of str, optional
        if provided compare these model statistics
    detailed_comparison : bool, optional
        if True return DataFrame containing comparison details,
        by default False which returns True if models are equivalent
        or False if they are not

    Returns
    -------
    bool or pd.DataFrame
        returns True if models are equivalent when detailed_comparison=True
        else returns DataFrame containing comparison details.
    """

    df = pd.DataFrame(columns=["model 0", "model 1"])
    so1 = []  # for storing series_original
    sv1 = []  # for storing series_validated
    ss1 = []  # for storing series

    for i, ml in enumerate([ml1, ml2]):
        counter = 0  # for counting stress time series
        df.loc["name:", f"model {i}"] = ml.name

        for k in ml.settings.keys():
            df.loc[f"- settings: {k}", f"model {i}"] = ml.settings.get(k)

        if i == 0:
            oso = (
                ml.oseries.series_original
                if PASTAS_LEQ_022
                else ml.oseries._series_original
            )
            df.loc["oseries: series_original", f"model {i}"] = True

            if PASTAS_LEQ_022:
                osv = ml.oseries.series_validated
                df.loc["oseries: series_validated", f"model {i}"] = True

            oss = ml.oseries.series
            df.loc["oseries: series_series", f"model {i}"] = True

        elif i == 1:
            try:
                assert_series_equal(
                    oso,
                    ml.oseries.series_original
                    if PASTAS_LEQ_022
                    else ml.oseries._series_original,
                )
                compare_oso = True
            except (ValueError, AssertionError):
                # series are not identical in length or index does not match
                compare_oso = False

            if PASTAS_LEQ_022:
                try:
                    assert_series_equal(osv, ml.oseries.series_validated)
                    compare_osv = True
                except (ValueError, AssertionError):
                    # series are not identical in length or index does not match
                    compare_osv = False

            try:
                assert_series_equal(oss, ml.oseries.series)
                compare_oss = True
            except (ValueError, AssertionError):
                # series are not identical in length or index does not match
                compare_oss = False

            df.loc["oseries: series_original", f"model {i}"] = compare_oso

            if PASTAS_LEQ_022:
                df.loc["oseries: series_validated", f"model {i}"] = compare_osv

            df.loc["oseries: series_series", f"model {i}"] = compare_oss

        for sm_name, sm in ml.stressmodels.items():
            df.loc[f"stressmodel: '{sm_name}'"] = sm_name
            df.loc["- rfunc"] = sm.rfunc._name if sm.rfunc is not None else "NA"

            if sm._name == "RechargeModel":
                stresses = [sm.prec, sm.evap]
            else:
                stresses = sm.stress

            for ts in stresses:
                df.loc[f"- time series: '{ts.name}'"] = ts.name
                for tsk in ts.settings.keys():
                    df.loc[
                        f"  - {ts.name} settings: {tsk}", f"model {i}"
                    ] = ts.settings[tsk]

                if i == 0:
                    if PASTAS_LEQ_022:
                        so1.append(ts.series_original.copy())
                        sv1.append(ts.series_validated.copy())
                    else:
                        so1.append(ts._series_original.copy())

                    ss1.append(ts.series.copy())

                    if PASTAS_LEQ_022:
                        df.loc[f"  - {ts.name}: series_validated"] = True

                    df.loc[f"  - {ts.name}: series_original"] = True
                    df.loc[f"  - {ts.name}: series"] = True

                elif i == 1:
                    # ValueError if series cannot be compared,
                    # set result to False
                    try:
                        assert_series_equal(
                            so1[counter],
                            ts.series_original
                            if PASTAS_LEQ_022
                            else ts._series_original,
                        )
                        compare_so1 = True
                    except (ValueError, AssertionError):
                        compare_so1 = False

                    if PASTAS_LEQ_022:
                        try:
                            assert_series_equal(sv1[counter], ts.series_validated)
                            compare_sv1 = True
                        except (ValueError, AssertionError):
                            compare_sv1 = False

                    try:
                        assert_series_equal(ss1[counter], ts.series)
                        compare_ss1 = True
                    except (ValueError, AssertionError):
                        compare_ss1 = False
                    df.loc[f"  - {ts.name}: series_original"] = compare_so1
                    if PASTAS_LEQ_022:
                        df.loc[f"  - {ts.name}: series_validated"] = compare_sv1
                    df.loc[f"  - {ts.name}: series"] = compare_ss1

                counter += 1

        for p in ml.parameters.index:
            df.loc[f"param: {p} (init)", f"model {i}"] = ml.parameters.loc[p, "initial"]
            df.loc[f"param: {p} (opt)", f"model {i}"] = ml.parameters.loc[p, "optimal"]

        if stats:
            stats_df = ml.stats.summary(stats=stats)
            for s in stats:
                df.loc[f"stats: {s}", f"model {i}"] = stats_df.loc[s, "Value"]

    # compare
    df["comparison"] = df.iloc[:, 0] == df.iloc[:, 1]

    # isclose for params
    param_mask = df.index.str.startswith("param: ")
    df.loc[param_mask, "comparison"] = np.isclose(
        df.loc[param_mask, "model 0"].astype(float).values,
        df.loc[param_mask, "model 1"].astype(float).values,
    )

    # ensure NaN == NaN is not counted as a difference
    nanmask = df.iloc[:, 0].isna() & df.iloc[:, 1].isna()
    df.loc[nanmask, "comparison"] = True

    # for stats comparison must be almost_equal
    if stats:
        stats_idx = [f"stats: {s}" for s in stats]
        b = np.isclose(
            df.loc[stats_idx, "model 0"].astype(float).values,
            df.loc[stats_idx, "model 1"].astype(float).values,
        )
        df.loc[stats_idx, "comparison"] = b

    if detailed_comparison:
        return df
    else:
        return df["comparison"].iloc[1:].all()  # ignore name difference


def copy_database(
    conn1,
    conn2,
    libraries: Optional[List[str]] = None,
    overwrite: bool = False,
    progressbar: bool = False,
) -> None:
    """Copy libraries from one database to another.

    Parameters
    ----------
    conn1 : pastastore.*Connector
        source Connector containing link to current database containing data
    conn2 : pastastore.*Connector
        destination Connector with link to database to which you want to copy
    libraries : Optional[List[str]], optional
        list of str containing names of libraries to copy, by default None,
        which copies all libraries: ['oseries', 'stresses', 'models']
    overwrite : bool, optional
        overwrite data in destination database, by default False
    progressbar : bool, optional
        show progressbars, by default False

    Raises
    ------
    ValueError
        if library name is not understood
    """
    if libraries is None:
        libraries = ["oseries", "stresses", "models"]

    for lib in libraries:
        if lib == "oseries":
            for name in (
                tqdm(conn1.oseries_names, desc="copying oseries")
                if progressbar
                else conn1.oseries_names
            ):
                o, meta = conn1.get_oseries(name, return_metadata=True)
                conn2.add_oseries(o, name, metadata=meta, overwrite=overwrite)
        elif lib == "stresses":
            for name in (
                tqdm(conn1.stresses_names, desc="copying stresses")
                if progressbar
                else conn1.stresses_names
            ):
                s, meta = conn1.get_stresses(name, return_metadata=True)
                conn2.add_stress(
                    s,
                    name,
                    kind=meta["kind"],
                    metadata=meta,
                    overwrite=overwrite,
                )
        elif lib == "models":
            for name in (
                tqdm(conn1.model_names, desc="copying models")
                if progressbar
                else conn1.model_names
            ):
                mldict = conn1.get_models(name, return_dict=True)
                conn2.add_model(mldict, overwrite=overwrite)
        else:
            raise ValueError(f"Library name '{lib}' not recognized!")


def frontiers_checks(
    pstore,
    modelnames: Optional[List[str]] = None,
    oseries: Optional[List[str]] = None,
    check1_rsq: bool = True,
    check1_threshold: float = 0.7,
    check2_autocor: bool = True,
    check2_test: str = "runs",
    check2_pvalue: float = 0.05,
    check3_tmem: bool = True,
    check3_cutoff: float = 0.95,
    check4_gain: bool = True,
    check5_parambounds: bool = False,
    csv_dir: Optional[str] = None,
) -> pd.DataFrame:  # pragma: no cover
    """Check models in a PastaStore to see if they pass reliability criteria.

    The reliability criteria are taken from Brakenhoff et al. 2022 [bra_2022]_.
    These criteria were applied in a region with recharge, river levels and
    pumping wells as stresses. This is by no means an exhaustive list of
    reliability criteria but might serve as a reasonable starting point for
    model diagnostic checking.

    Parameters
    ----------
    pstore : pastastore.PastaStore
        reference to a PastaStore
    modelnames : list of str, optional
        list of model names to consider, if None checks 'oseries', if both are
        None, all stored models will be checked
    oseries :  list of str, optional
        list of oseries to consider, corresponding models will be picked up
        from pastastore. If None, uses all stored models are checked.
    check1 : bool, optional
        check if model fit is above a threshold of the coefficient
        of determination $R^2$ , by default True
    check1_threshold : float, optional
        threshold of the $R^2$ fit statistic, by default 0.7
    check2 : bool, optional
        check if the noise of the model has autocorrelation with
        statistical test, by default True
    check2_test : str, optional
        statistical test for autocorrelation. Available options are Runs
        test "runs", Stoffer-Toloi "stoffer" or "both", by default "runs"
    check2_pvalue : float, optional
        p-value for the statistical test to define the confindence
        interval, by default 0.05
    check3 : bool, optional
        check if the length of the response time is within the
        calibration period, by default True
    check3_cutoff : float, optional
        the cutoff of the response time, by default 0.95
    check4 : bool, optional
        check if the uncertainty of the gain, by default True
    check5 : bool, optional
        check if parameters hit parameter bounds, by default False
    csv_dir : string, optional
        directory to store CSV file with overview of checks for every
        model, by default None which will not store results

    Returns
    -------
    df : pandas.DataFrame
        dataFrame with all models and whether or not they pass
        the reliability checks


    References
    ----------
    .. [bra_2022]
    Brakenhoff, D.A., Vonk M.A., Collenteur, R.A., van Baar, M., Bakker, M.:
    Application of Time Series Analysis to Estimate Drawdown From Multiple Well
    Fields. Front. Earth Sci., 14 June 2022 doi:10.3389/feart.2022.907609
    """

    df = pd.DataFrame(columns=["all_checks_passed"])

    if modelnames is not None:
        models = modelnames
        if oseries is not None:
            print(
                "Warning! Both 'modelnames' and 'oseries' provided,"
                " only using 'modelnames'!"
            )
    elif oseries is not None:
        models = []
        for o in oseries:
            models += pstore.oseries_models[o]
    else:
        models = pstore.model_names

    for mlnam in tqdm(models, desc="Running model diagnostics"):
        ml = pstore.get_models(mlnam)

        if ml.parameters["optimal"].hasnans:
            print(f"Warning! Skipping model '{mlnam}' because " "it is not solved!")
            continue

        checks = pd.DataFrame(columns=["stat", "threshold", "units", "check_passed"])

        # Check 1 - Fit Statistic
        if check1_rsq:
            rsq = ml.stats.rsq()
            check_rsq_passed = rsq >= check1_threshold
            checks.loc["rsq >= threshold", :] = (
                rsq,
                check1_threshold,
                "-",
                check_rsq_passed,
            )

        # Check 2 - Autocorrelation Noise
        if check2_autocor:
            noise = ml.noise()
            if noise is None:
                noise = ml.residuals()
                print(
                    "Warning! Checking autocorrelation on the residuals not the noise!"
                )
            if check2_test == "runs" or check2_test == "both":
                _, p_runs = runs_test(noise.iloc[1:])
                if p_runs > check2_pvalue:  # No autocorrelation
                    check_runs_acf_passed = True
                else:  # Significant autocorrelation
                    check_runs_acf_passed = False
                checks.loc["ACF: Runs test", :] = (
                    p_runs,
                    check2_pvalue,
                    "-",
                    check_runs_acf_passed,
                )
            if check2_test == "stoffer" or check2_test == "both":
                _, p_stoffer = stoffer_toloi(
                    noise.iloc[1:], snap_to_equidistant_timestamps=True
                )
                if p_stoffer > check2_pvalue:
                    check_st_acf_passed = True
                else:
                    check_st_acf_passed = False
                checks.loc["ACF: Stoffer-Toloi test", :] = (
                    p_stoffer,
                    check2_pvalue,
                    "-",
                    check_st_acf_passed,
                )

        # Check 3 - Response Time
        if check3_tmem:
            len_oseries_calib = (ml.settings["tmax"] - ml.settings["tmin"]).days
            for sm_name, sm in ml.stressmodels.items():
                if sm._name == "WellModel":
                    nwells = sm.distances.index.size
                    for iw in range(nwells):
                        p = sm.get_parameters(model=ml, istress=iw)
                        t = sm.rfunc.get_t(p, dt=1, cutoff=0.999)
                        step = sm.rfunc.step(p, cutoff=0.999) / sm.rfunc.gain(p)
                        tmem = np.interp(check3_cutoff, step, t)
                        check_tmem_passed = tmem < len_oseries_calib / 2
                        idxlbl = (
                            f"calib_period > 2*t_mem_95%: "
                            f"{sm_name}-{iw:02g} ({sm.distances.index[iw]})"
                        )
                        checks.loc[idxlbl, :] = (
                            tmem,
                            len_oseries_calib,
                            "days",
                            check_tmem_passed,
                        )
                else:
                    tmem = ml.get_response_tmax(sm_name)
                    if tmem is None:  # no rfunc in stressmodel
                        tmem = 0
                    check_tmem_passed = tmem < len_oseries_calib / 2
                    checks.loc[f"calib_period > 2*t_mem_95%: {sm_name}", :] = (
                        tmem,
                        len_oseries_calib,
                        "days",
                        check_tmem_passed,
                    )

        # Check 4 - Uncertainty Gain
        if check4_gain:
            for sm_name, sm in ml.stressmodels.items():
                if sm._name == "WellModel":
                    for iw in range(sm.distances.index.size):
                        p = sm.get_parameters(model=ml, istress=iw)
                        gain = sm.rfunc.gain(p)
                        gain_std = np.sqrt(sm.variance_gain(model=ml, istress=iw))
                        if gain_std is None:
                            gain_std = np.nan
                            check_gain_passed = pd.NA
                        elif np.isnan(gain_std):
                            check_gain_passed = pd.NA
                        else:
                            check_gain_passed = np.abs(gain) > 2 * gain_std
                        checks.loc[
                            f"gain > 2*std: {sm_name}-{iw:02g} ({sm.distances.index[iw]})",
                            :,
                        ] = (
                            gain,
                            2 * gain_std,
                            "(unit head)/(unit well stress)",
                            check_gain_passed,
                        )
                    continue
                elif sm._name == "LinearTrend":
                    gain = ml.parameters.loc[f"{sm_name}_a", "optimal"]
                    gain_std = ml.parameters.loc[f"{sm_name}_a", "stderr"]
                elif sm._name == "StepModel":
                    gain = ml.parameters.loc[f"{sm_name}_d", "optimal"]
                    gain_std = ml.parameters.loc[f"{sm_name}_d", "stderr"]
                else:
                    gain = ml.parameters.loc[f"{sm_name}_A", "optimal"]
                    gain_std = ml.parameters.loc[f"{sm_name}_A", "stderr"]

                if gain_std is None:
                    gain_std = np.nan
                    check_gain_passed = pd.NA
                elif np.isnan(gain_std):
                    check_gain_passed = pd.NA
                else:
                    check_gain_passed = np.abs(gain) > 2 * gain_std
                checks.loc[f"gain > 2*std: {sm_name}", :] = (
                    gain,
                    2 * gain_std,
                    "(unit head)/(unit well stress)",
                    check_gain_passed,
                )

        # Check 5 - Parameter Bounds
        if check5_parambounds:
            upper, lower = ml._check_parameters_bounds()
            for param in ml.parameters.index:
                bounds = (
                    ml.parameters.loc[param, "pmin"],
                    ml.parameters.loc[param, "pmax"],
                )
                b = ~(upper.loc[param] or lower.loc[param])

                checks.loc[f"Parameter bounds: {param}", :] = (
                    ml.parameters.loc[param, "optimal"],
                    bounds,
                    "_",
                    b,
                )

        df.loc[mlnam, "all_checks_passed"] = checks["check_passed"].all()
        df.loc[mlnam, checks.index] = checks.loc[:, "check_passed"]

        if csv_dir:
            checks.to_csv(f"{csv_dir}/checks_{ml.name}.csv", na_rep="NaN")

    return df


def frontiers_aic_select(
    pstore,
    modelnames: Optional[List[str]] = None,
    oseries: Optional[List[str]] = None,
    full_output: bool = False,
) -> pd.DataFrame:  # pragma: no cover
    """Select the best model structure based on the minimum AIC.

    As proposed by Brakenhoff et al. 2022 [bra_2022]_.

    Parameters
    ----------
    pstore : pastastore.PastaStore
        reference to a PastaStore
    modelnames : list of str
        list of model names (that pass reliability criteria)
    oseries : list of oseries
        list of locations for which to select models, note that this uses all
        models associated with a specific location.
    full_output : bool, optional
        if set to True, returns a DataFrame including all models per location
        and their AIC values

    Returns
    -------
    pandas.DataFrame
        DataFrame with selected best model per location based on the AIC, or a
        DataFrame containing statistics for each of the models per location

    References
    ----------
    .. [bra_2022] Brakenhoff, D.A., Vonk M.A., Collenteur, R.A., van Baar, M.,
    Bakker, M.: Application of Time Series Analysis to Estimate Drawdown From
    Multiple Well Fields. Front. Earth Sci., 14 June 2022
    doi:10.3389/feart.2022.907609
    """

    if modelnames is None and oseries is None:
        modelnames = pstore.model_names
    elif modelnames is None and oseries is not None:
        modelnames = []
        for o in oseries:
            modelnames += pstore.oseries_models[o]
    elif oseries is not None:
        print(
            "Warning! Both 'modelnames' and 'oseries' provided, "
            "using only 'modelnames'"
        )

    # Dataframe of models with corresponding oseries
    df = pstore.get_model_timeseries_names(modelnames, progressbar=False).loc[
        :, ["oseries"]
    ]
    # AIC of models
    aic = pstore.get_statistics(["aic"], modelnames, progressbar=True)
    if full_output:
        # group models per location and obtain the AIC identify model
        # with lowest AIC per location
        collect = []
        gr = df.join(aic).groupby("oseries")
        for o, idf in gr:
            idf.index.name = "modelname"
            idf = (
                idf.sort_values("aic").reset_index().set_index(["oseries", "modelname"])
            )
            idf = idf.rename(columns={"aic": "AIC"})
            idf["dAIC"] = idf["AIC"] - idf["AIC"].min()
            idf = idf.replace(0.0, np.nan)
            collect.append(idf)
        return pd.concat(collect, axis=0)
    else:
        return (
            df.join(aic).groupby("oseries").idxmin().rename(columns={"aic": "min_aic"})
        )



import matplotlib.pyplot as plt
import pytest
from conftest import requires_pkg
from pytest_dependency import depends

# %% plots


def test_plot_oseries(pstore):
    ax = pstore.plots.oseries()
    plt.close(ax.figure)


def test_plot_stresses(pstore):
    ax = pstore.plots.stresses()
    plt.close(ax.figure)


def test_plot_stresses_availability(pstore):
    ax = pstore.plots.data_availability("stresses", kind="prec", set_yticks=True)
    plt.close(ax.figure)


@pytest.mark.dependency()
def test_cumulative_hist(request, pstore):
    ml1 = pstore.create_model("oseries1")
    pstore.add_model(ml1)
    ml2 = pstore.create_model("oseries2")
    pstore.add_model(ml2)
    ax = pstore.plots.cumulative_hist()
    plt.close(ax.figure)


# %% maps


@pytest.mark.bgmap
def test_map_oseries_w_bgmap(pstore):
    ax = pstore.maps.oseries()
    # only test bgmap once for pas
    if pstore.type == "pas":
        pstore.maps.add_background_map(ax)
    plt.close(ax.figure)


@requires_pkg("adjustText")
def test_map_stresses(pstore):
    ax = pstore.maps.stresses(kind="prec", adjust=True)
    plt.close(ax.figure)


def test_map_stresslinks(pstore):
    ml = pstore.create_model("oseries1", modelname="ml1")
    pstore.add_model(ml)
    ax = pstore.maps.stresslinks()
    plt.close(ax.figure)


@pytest.mark.dependency()
def test_map_models(request, pstore):
    ax = pstore.maps.models()
    plt.close(ax.figure)


@pytest.mark.dependency()
def test_map_model(request, pstore):
    depends(request, [f"test_map_models[{pstore.type}]"])
    ax = pstore.maps.model("oseries1")
    plt.close(ax.figure)


@pytest.mark.dependency()
def test_map_modelstat(request, pstore):
    ax = pstore.maps.modelstat("evp")
    plt.close(ax.figure)


@pytest.mark.dependency()
def test_list_ctx_providers(request, pstore):
    pstore.maps._list_contextily_providers()



"""This module contains all the plotting methods for PastaStore.

Pastastore comes with a number helpful plotting methods to quickly
visualize time series or the locations of the time series contained in the
store. Plotting time series or data availability is available through the
`plots` attribute of the PastaStore object. Plotting locations of time series
or model statistics on maps is available through the `maps` attribute.
For example, if we have a :class:`pastastore.PastaStore` called `pstore`
linking to an existing database, the plot and map methods are available as
follows::

    pstore.plots.oseries()

    ax = pstore.maps.oseries()
    pstore.maps.add_background_map(ax)  # for adding a background map
"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pastas as ps
from matplotlib import patheffects
from matplotlib.collections import LineCollection
from matplotlib.colors import BoundaryNorm, LogNorm
from matplotlib.lines import Line2D
from mpl_toolkits.axes_grid1 import make_axes_locatable


class Plots:
    """Plot class for Pastastore.

    Allows plotting of time series and data availability.
    """

    def __init__(self, pstore):
        """Initialize Plots class for Pastastore.

        Parameters
        ----------
        pstore : pastastore.Pastastore
            Pastastore object
        """
        self.pstore = pstore

    def _timeseries(
        self,
        libname,
        names=None,
        ax=None,
        split=False,
        figsize=(10, 5),
        progressbar=True,
        **kwargs,
    ):
        """Internal method to plot time series from pastastore.

        Parameters
        ----------
        libname : str
            name of the library to obtain time series from (oseries
            or stresses)
        names : list of str, optional
            list of time series names to plot, by default None
        ax : matplotlib.Axes, optional
            pass axes object to plot on existing axes, by default None,
            which creates a new figure
        split : bool, optional
            create a separate subplot for each time series, by default False.
            A maximum of 20 time series is supported when split=True.
        figsize : tuple, optional
            figure size, by default (10, 5)
        progressbar : bool, optional
            show progressbar when loading time series from store,
            by default True

        Returns
        -------
        ax : matplotlib.Axes
            axes handle

        Raises
        ------
        ValueError
            split=True is only supported if there are less than 20 time series
            to plot.
        """
        names = self.pstore.conn._parse_names(names, libname)

        if len(names) > 20 and split:
            raise ValueError(
                "More than 20 time series leads to too many "
                "subplots, set split=False."
            )

        if ax is None:
            if split:
                _, axes = plt.subplots(len(names), 1, sharex=True, figsize=figsize)
            else:
                _, axes = plt.subplots(1, 1, figsize=figsize)
        else:
            axes = ax

        tsdict = self.pstore.conn._get_series(
            libname, names, progressbar=progressbar, squeeze=False
        )
        for i, (n, ts) in enumerate(tsdict.items()):
            if split and ax is None:
                iax = axes[i]
            elif ax is None:
                iax = axes
            else:
                iax = ax
            iax.plot(ts.index, ts.squeeze(), label=n, **kwargs)
            if split:
                iax.legend(loc="best", fontsize="x-small")

        if not split:
            axes.legend(loc=(0, 1), frameon=False, ncol=7, fontsize="x-small")

        return axes

    def oseries(self, names=None, ax=None, split=False, figsize=(10, 5), **kwargs):
        """Plot oseries.

        Parameters
        ----------
        names : list of str, optional
            list of oseries names to plot, by default None, which loads
            all oseries from store
        ax : matplotlib.Axes, optional
            pass axes object to plot oseries on existing figure,
            by default None, in which case a new figure is created
        split : bool, optional
            create a separate subplot for each time series, by default False.
            A maximum of 20 time series is supported when split=True.
        figsize : tuple, optional
            figure size, by default (10, 5)

        Returns
        -------
        ax : matplotlib.Axes
            axes handle
        """
        return self._timeseries(
            "oseries",
            names=names,
            ax=ax,
            split=split,
            figsize=figsize,
            **kwargs,
        )

    def stresses(
        self,
        names=None,
        kind=None,
        ax=None,
        split=False,
        figsize=(10, 5),
        **kwargs,
    ):
        """Plot stresses.

        Parameters
        ----------
        names : list of str, optional
            list of oseries names to plot, by default None, which loads
            all oseries from store
        kind : str, optional
            only plot stresses of a certain kind, by default None, which
            includes all stresses
        ax : matplotlib.Axes, optional
            pass axes object to plot oseries on existing figure,
            by default None, in which case a new figure is created
        split : bool, optional
            create a separate subplot for each time series, by default False.
            A maximum of 20 time series is supported when split=True.
        figsize : tuple, optional
            figure size, by default (10, 5)

        Returns
        -------
        ax : matplotlib.Axes
            axes handle
        """
        names = self.pstore.conn._parse_names(names, "stresses")
        masknames = self.pstore.stresses.index.isin(names)
        stresses = self.pstore.stresses.loc[masknames]

        if kind:
            mask = stresses["kind"] == kind
            names = stresses.loc[mask].index.to_list()

        return self._timeseries(
            "stresses",
            names=names,
            ax=ax,
            split=split,
            figsize=figsize,
            **kwargs,
        )

    def data_availability(
        self,
        libname,
        names=None,
        kind=None,
        intervals=None,
        ignore=("second", "minute", "14 days"),
        ax=None,
        cax=None,
        normtype="log",
        cmap="viridis_r",
        set_yticks=False,
        figsize=(10, 8),
        progressbar=True,
        dropna=True,
        **kwargs,
    ):
        """Plot the data-availability for multiple time series in pastastore.

        Parameters
        ----------
        libname : str
            name of library to get time series from (oseries or stresses)
        names : list, optional
            specify names in a list to plot data availability for certain
            time series
        kind : str, optional
            if library is stresses, kind can be specified to obtain only
            stresses of a specific kind
        intervals: dict, optional
            A dict with frequencies as keys and number of seconds as values
        ignore : list, optional
            A list with frequencies in intervals to ignore
        ax: matplotlib Axes, optional
            pass axes object to plot data availability on existing figure. by
            default None, in which case a new figure is created
        cax: matplotlib Axes, optional
            pass object axes to plot the colorbar on. by default None, which
            gives default Maptlotlib behavior
        normtype : str, optional
            Determines the type of color normalisations, default is 'log'
        cmap : str, optional
            A reference to a matplotlib colormap
        set_yticks : bool, optional
            Set the names of the series as yticks
        figsize : tuple, optional
            The size of the new figure in inches (h,v)
        progressbar : bool
            Show progressbar
        dropna : bool
            Do not show NaNs as available data
        kwargs : dict, optional
            Extra arguments are passed to matplotlib.pyplot.subplots()

        Returns
        -------
        ax : matplotlib Axes
            The axes in which the data-availability is plotted
        """
        names = self.pstore.conn._parse_names(names, libname)

        if libname == "stresses":
            masknames = self.pstore.stresses.index.isin(names)
            stresses = self.pstore.stresses.loc[masknames]
            if kind:
                mask = stresses["kind"] == kind
                names = stresses.loc[mask].index.to_list()

        series = self.pstore.conn._get_series(
            libname, names, progressbar=progressbar, squeeze=False
        ).values()

        ax = self._data_availability(
            series,
            names=names,
            intervals=intervals,
            ignore=ignore,
            ax=ax,
            cax=cax,
            normtype=normtype,
            cmap=cmap,
            set_yticks=set_yticks,
            figsize=figsize,
            dropna=dropna,
            **kwargs,
        )
        return ax

    @staticmethod
    def _data_availability(
        series,
        names=None,
        intervals=None,
        ignore=("second", "minute", "14 days"),
        ax=None,
        cax=None,
        normtype="log",
        cmap="viridis_r",
        set_yticks=False,
        figsize=(10, 8),
        dropna=True,
        **kwargs,
    ):
        """Plot the data-availability for a list of time series.

        Parameters
        ----------
        libname : list of pandas.Series
            list of series to plot data availability for
        names : list, optional
            specify names of series, default is None in which case names
            will be taken from series themselves.
        kind : str, optional
            if library is stresses, kind can be specified to obtain only
            stresses of a specific kind
        intervals: dict, optional
            A dict with frequencies as keys and number of seconds as values
        ignore : list, optional
            A list with frequencies in intervals to ignore
        ax: matplotlib Axes, optional
            pass axes object to plot data availability on existing figure. by
            default None, in which case a new figure is created
        cax: matplotlib Axes, optional
            pass object axes to plot the colorbar on. by default None, which
            gives default Maptlotlib behavior
        normtype : str, optional
            Determines the type of color normalisations, default is 'log'
        cmap : str, optional
            A reference to a matplotlib colormap
        set_yticks : bool, optional
            Set the names of the series as yticks
        figsize : tuple, optional
            The size of the new figure in inches (h,v)
        progressbar : bool
            Show progressbar
        dropna : bool
            Do not show NaNs as available data
        kwargs : dict, optional
            Extra arguments are passed to matplotlib.pyplot.subplots()

        Returns
        -------
        ax : matplotlib Axes
            The axes in which the data-availability is plotted
        """
        # a good colormap is cmap='RdYlGn_r' or 'cubehelix'
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize, **kwargs)
        else:
            fig = ax.get_figure()

        ax.invert_yaxis()
        if intervals is None:
            intervals = {
                "second": 1,
                "minute": 60,
                "hour": 60 * 60,
                "day": 60 * 60 * 24,
                "week": 60 * 60 * 24 * 7,
                "14 days": 60 * 60 * 24 * 14,
                "month": 60 * 60 * 24 * 31,
                "quarter": 60 * 60 * 24 * 31 * 4,
                "year": 60 * 60 * 24 * 366,
            }
            for i in ignore:
                if i in intervals:
                    intervals.pop(i)

        bounds = np.array([intervals[i] for i in intervals])
        bounds = bounds.astype(float) * (10**9)
        labels = intervals.keys()
        if normtype == "log":
            norm = LogNorm(vmin=bounds[0], vmax=bounds[-1])
        else:
            norm = BoundaryNorm(boundaries=bounds, ncolors=256)
        cmap = plt.get_cmap(cmap, 256)
        cmap.set_over((1.0, 1.0, 1.0))

        for i, s in enumerate(series):
            if not s.empty:
                if dropna:
                    s = s.dropna()
                pc = ax.pcolormesh(
                    s.index,
                    [i, i + 1],
                    [np.diff(s.index).astype(float)],
                    norm=norm,
                    cmap=cmap,
                    linewidth=0,
                    rasterized=True,
                )

        # make a colorbar in an ax on the
        # right side, then set the current axes to ax again
        cb = fig.colorbar(pc, ax=ax, cax=cax, extend="both")
        cb.set_ticks(bounds)
        cb.ax.set_yticklabels(labels)
        cb.ax.minorticks_off()

        if set_yticks:
            ax.set_yticks(np.arange(0.5, len(series) + 0.5), minor=False)
            ax.set_yticks(np.arange(0, len(series) + 1), minor=True)
            if names is None:
                names = [s.name for s in series]
            ax.set_yticklabels(names)

            for tick in ax.yaxis.get_major_ticks():  # don't show major ytick marker
                tick.tick1line.set_visible(False)

            ax.grid(True, which="minor", axis="y")
            ax.grid(True, which="major", axis="x")

        else:
            ax.set_ylabel("Timeseries (-)")
            ax.grid(True, which="both")
            ax.grid(True, which="both")

        return ax

    def cumulative_hist(
        self,
        statistic="rsq",
        modelnames=None,
        extend=False,
        ax=None,
        figsize=(6, 6),
        label=None,
        legend=True,
    ):
        """Plot a cumulative step histogram for a model statistic.

        Parameters
        ----------
        statistic: str
            name of the statistic, e.g. "evp" or "rmse", by default "rsq"
        modelnames: list of str, optional
            modelnames to plot statistic for, by default None, which
            uses all models in the store
        extend: bool, optional
            force extend the stats Series with a dummy value to move the
            horizontal line outside figure bounds. If True the results
            are skewed a bit, especially if number of models is low.
        ax: matplotlib.Axes, optional
            axes to plot histogram, by default None which creates an Axes
        figsize: tuple, optional
            figure size, by default (6,6)
        label: str, optional
            label for the legend, by default None, which shows the number
            of models
        legend: bool, optional
            show legend, by default True

        Returns
        -------
        ax : matplotlib Axes
            The axes in which the cumulative histogram is plotted
        """

        statsdf = self.pstore.get_statistics(
            [statistic], modelnames=modelnames, progressbar=False
        )

        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=figsize)
            ax.set_xticks(np.linspace(0, 1, 11))
            ax.set_xlim(0, 1)
            ax.set_ylabel(statistic)
            ax.set_xlabel("Density")
            ax.set_title("Cumulative Step Histogram")
        if statistic == "evp":
            ax.set_yticks(np.linspace(0, 100, 11))
            if extend:
                statsdf = statsdf.append(pd.Series(100, index=["dummy"]))
                ax.set_ylim(0, 100)
            else:
                ax.set_ylim(0, statsdf.max())
        elif statistic in ("rsq", "nse", "kge_2012"):
            ax.set_yticks(np.linspace(0, 1, 11))
            if extend:
                statsdf = statsdf.append(pd.Series(1, index=["dummy"]))
                statsdf[statsdf < 0] = 0
                ax.set_ylim(0, 1)
            else:
                ax.set_ylim(0, statsdf.max())
        elif statistic in ("aic", "bic"):
            ax.set_ylim(statsdf.min(), statsdf.max())
        else:
            if extend:
                statsdf = statsdf.append(pd.Series(0, index=["dummy"]))
            ax.set_ylim(0, statsdf.max())

        if label is None:
            if extend:
                label = f"No. Models = {len(statsdf)-1}"
            else:
                label = f"No. Models = {len(statsdf)}"

        statsdf.hist(
            ax=ax,
            bins=len(statsdf),
            density=True,
            cumulative=True,
            histtype="step",
            orientation="horizontal",
            label=label,
        )

        if legend:
            ax.legend(loc=4)

        return ax


class Maps:
    """Map Class for PastaStore.

    Allows plotting locations and model statistics on maps.

    Usage
    -----
    Example usage of the maps methods: :

    >> > ax = pstore.maps.oseries()  # plot oseries locations
    >> > pstore.maps.add_background_map(ax)  # add background map
    """

    def __init__(self, pstore):
        """Initialize Plots class for Pastastore.

        Parameters
        ----------
        pstore: pastastore.Pastastore
            Pastastore object
        """
        self.pstore = pstore

    def stresses(
        self,
        names=None,
        kind=None,
        labels=True,
        adjust=False,
        figsize=(10, 8),
        backgroundmap=False,
        **kwargs,
    ):
        """Plot stresses locations on map.

        Parameters
        ----------
        names : list of str, optional
            list of names to plot
        kind: str, optional
            if passed, only plot stresses of a specific kind, default is None
            which plots all stresses.
        labels: bool, optional
            label models, by default True
        adjust: bool, optional
            automated smart label placement using adjustText, by default False
        ax : matplotlib.Axes, optional
            axes handle, if not provided a new figure is created.
        figsize: tuple, optional
            figure size, by default(10, 8)
        backgroundmap: bool, optional
            if True, add background map (default CRS is EPSG:28992) with default tiles
            by OpenStreetMap.Mapnik. Default option is False.

        Returns
        -------
        ax: matplotlib.Axes
            axes object

        See also
        --------
        self.add_background_map
        """
        if names is not None:
            df = self.pstore.stresses.loc[names]
        else:
            df = self.pstore.stresses

        if kind is not None:
            if isinstance(kind, str):
                kind = [kind]
            mask = df["kind"].isin(kind)
            stresses = df[mask]
        else:
            stresses = df

        mask0 = (stresses["x"] != 0.0) | (stresses["y"] != 0.0)

        if "c" in kwargs:
            c = kwargs.pop("c", None)
        else:
            c = stresses.loc[mask0, "kind"]
            kind_to_color = {k: f"C{i}" for i, k in enumerate(c.unique())}
            c = c.apply(lambda k: kind_to_color[k])

        r = self._plotmap_dataframe(stresses.loc[mask0], c=c, figsize=figsize, **kwargs)
        if "ax" in kwargs:
            ax = kwargs["ax"]
        else:
            ax = r
        if labels:
            self.add_labels(stresses, ax, adjust=adjust)

        if backgroundmap:
            self.add_background_map(ax)

        return ax

    def oseries(
        self,
        names=None,
        labels=True,
        adjust=False,
        figsize=(10, 8),
        backgroundmap=False,
        **kwargs,
    ):
        """Plot oseries locations on map.

        Parameters
        ----------
        names: list, optional
            oseries names, by default None which plots all oseries locations
        labels: bool, optional
            label models, by default True
        adjust: bool, optional
            automated smart label placement using adjustText, by default False
        figsize: tuple, optional
            figure size, by default(10, 8)
        backgroundmap: bool, optional
            if True, add background map (default CRS is EPSG:28992) with default tiles
            by OpenStreetMap.Mapnik. Default option is False.

        Returns
        -------
        ax: matplotlib.Axes
            axes object

        See also
        --------
        self.add_background_map
        """

        names = self.pstore.conn._parse_names(names, "oseries")
        oseries = self.pstore.oseries.loc[names]
        mask0 = (oseries["x"] != 0.0) | (oseries["y"] != 0.0)
        r = self._plotmap_dataframe(oseries.loc[mask0], figsize=figsize, **kwargs)
        if "ax" in kwargs:
            ax = kwargs["ax"]
        else:
            ax = r
        if labels:
            self.add_labels(oseries, ax, adjust=adjust)

        if backgroundmap:
            self.add_background_map(ax)

        return ax

    def models(
        self, labels=True, adjust=False, figsize=(10, 8), backgroundmap=False, **kwargs
    ):
        """Plot model locations on map.

        Parameters
        ----------
        labels: bool, optional
            label models, by default True
        adjust: bool, optional
            automated smart label placement using adjustText, by default False
        ax : matplotlib.Axes, optional
            axes handle, if not provided a new figure is created.
        figsize: tuple, optional
            figure size, by default(10, 8)
        backgroundmap: bool, optional
            if True, add background map (default CRS is EPSG:28992) with default tiles
            by OpenStreetMap.Mapnik. Default option is False.

        Returns
        -------
        ax: matplotlib.Axes
            axes object

        See also
        --------
        self.add_background_map
        """

        model_oseries = [
            self.pstore.get_models(m, return_dict=True)["oseries"]["name"]
            for m in self.pstore.model_names
        ]

        models = self.pstore.oseries.loc[model_oseries]
        models.index = self.pstore.model_names

        # mask out 0.0 coordinates
        mask0 = (models["x"] != 0.0) | (models["y"] != 0.0)
        r = self._plotmap_dataframe(models.loc[mask0], figsize=figsize, **kwargs)
        if "ax" in kwargs:
            ax = kwargs["ax"]
        else:
            ax = r
        if labels:
            self.add_labels(models, ax, adjust=adjust)

        if backgroundmap:
            self.add_background_map(ax)

        return ax

    def modelstat(
        self,
        statistic,
        modelnames=None,
        label=True,
        adjust=False,
        cmap="viridis",
        norm=None,
        vmin=None,
        vmax=None,
        figsize=(10, 8),
        backgroundmap=False,
        **kwargs,
    ):
        """Plot model statistic on map.

        Parameters
        ----------
        statistic: str
            name of the statistic, e.g. "evp" or "aic"
        modelnames : list of str, optional
            list of modelnames to include
        label: bool, optional
            label points, by default True
        adjust: bool, optional
            automated smart label placement using adjustText, by default False
        cmap: str or colormap, optional
            (name of) the colormap, by default "viridis"
        norm: norm, optional
            normalization for colorbar, by default None
        vmin: float, optional
            vmin for colorbar, by default None
        vmax: float, optional
            vmax for colorbar, by default None
        ax : matplotlib.Axes, optional
            axes handle, if not provided a new figure is created.
        figsize: tuple, optional
            figuresize, by default(10, 8)
        backgroundmap: bool, optional
            if True, add background map (default CRS is EPSG:28992) with default tiles
            by OpenStreetMap.Mapnik. Default option is False.

        Returns
        -------
        ax: matplotlib.Axes
            axes object

        See also
        --------
        self.add_background_map
        """
        statsdf = self.pstore.get_statistics(
            [statistic], modelnames=modelnames, progressbar=False
        ).to_frame()

        statsdf["oseries"] = [
            self.pstore.get_models(m, return_dict=True)["oseries"]["name"]
            for m in statsdf.index
        ]
        statsdf = statsdf.reset_index().set_index("oseries")
        df = statsdf.join(self.pstore.oseries, how="left")

        scatter_kwargs = {
            "cmap": cmap,
            "norm": norm,
            "vmin": vmin,
            "vmax": vmax,
            "edgecolors": "w",
            "linewidths": 0.7,
        }

        scatter_kwargs.update(kwargs)

        ax = self._plotmap_dataframe(
            df, column=statistic, figsize=figsize, **scatter_kwargs
        )
        if label:
            df.set_index("index", inplace=True)
            self.add_labels(df, ax, adjust=adjust)

        if backgroundmap:
            self.add_background_map(ax)

        return ax

    @staticmethod
    def _plotmap_dataframe(
        df,
        x="x",
        y="y",
        column=None,
        colorbar=True,
        ax=None,
        figsize=(10, 8),
        **kwargs,
    ):
        """Internal method for plotting dataframe with point locations.

        Can be called directly for more control over plot characteristics.

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame containing coordinates and data to plot, with
            index providing names for each location
        x : str, optional
            name of the column with x - coordinate data, by default "x"
        y : str, optional
            name of the column with y - coordinate data, by default "y"
        column : str, optional
            name of the column containing data used for determining the
            color of each point, by default None (all one color)
        colorbar : bool, optional
            show colorbar, only if column is provided, by default True
        ax : matplotlib Axes
            axes handle to plot dataframe, optional, default is None
            which creates a new figure
        figsize : tuple, optional
            figure size, by default(10, 8)
        **kwargs :
            dictionary containing keyword arguments for ax.scatter,
            by default None

        Returns
        -------
        ax : matplotlib.Axes
            axes object, returned if ax is None
        sc : scatter handle
            scatter plot handle, returned if ax is not None
        """

        if ax is None:
            return_scatter = False
            fig, ax = plt.subplots(figsize=figsize)
        else:
            return_scatter = True
            fig = ax.figure

        # set default size and marker if not passed
        if kwargs:
            s = kwargs.pop("s", 70)
            marker = kwargs.pop("marker", "o")
        else:
            s = 70
            marker = "o"
            kwargs = {}

        # if column is passed for coloring pts
        if column:
            c = df[column]
            if "cmap" not in kwargs:
                kwargs["cmap"] = "viridis"
        else:
            c = kwargs.pop("c", None)

        sc = ax.scatter(df[x], df[y], marker=marker, s=s, c=c, **kwargs)
        # add colorbar
        if column and colorbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="3%", pad=0.05)
            cbar = fig.colorbar(sc, ax=ax, cax=cax)
            cbar.set_label(column)

        # set axes properties
        ax.set_xlabel("x")
        ax.set_ylabel("y")

        for label in ax.get_yticklabels():
            label.set_rotation(90)
            label.set_verticalalignment("center")

        fig.tight_layout()
        if return_scatter:
            return sc
        else:
            return ax

    def model(
        self,
        ml,
        label=True,
        metadata_source="model",
        offset=0.0,
        ax=None,
        figsize=(10, 10),
        backgroundmap=False,
    ):
        """Plot oseries and stresses from one model on a map.

        Parameters
        ----------
        ml: str or pastas.Model
            pastas model or name of pastas model to plot on map
        label: bool, optional, default is True
            add labels to points on map
        metadata_source: str, optional
            whether to obtain metadata from model Timeseries or from
            metadata in pastastore("store"), default is "model"
        offset : float, optional
            add offset to current extent of model time series, useful
            for zooming out around models
        ax : matplotlib.Axes, optional
            axes handle, if not provided a new figure is created.
        figsize: tuple, optional
            figsize, default is (10, 10)
        backgroundmap: bool, optional
            if True, add background map (default CRS is EPSG:28992) with default tiles
            by OpenStreetMap.Mapnik. Default option is False.

        Returns
        -------
        ax: axes object
            axis handle of the resulting figure

        See also
        --------
        self.add_background_map
        """
        if isinstance(ml, str):
            ml = self.pstore.get_models(ml)
        elif not isinstance(ml, ps.Model):
            raise TypeError("Pass model name as string or pastas.Model!")

        stresses = pd.DataFrame(columns=["x", "y", "stressmodel", "color"])
        count = 0
        for name, sm in ml.stressmodels.items():
            for istress in sm.stress:
                if metadata_source == "model":
                    xi = istress.metadata["x"]
                    yi = istress.metadata["y"]
                elif metadata_source == "store":
                    imeta = self.pstore.get_metadata(
                        "stresses", istress.name, as_frame=False
                    )
                    xi = imeta.pop("x", np.nan)
                    yi = imeta.pop("y", np.nan)
                else:
                    raise ValueError(
                        "metadata_source must be either " "'model' or 'store'!"
                    )
                if np.isnan(xi) or np.isnan(yi):
                    print(f"No x,y-data for {istress.name}!")
                    continue
                if xi == 0.0 or yi == 0.0:
                    print(f"x,y-data is 0.0 for {istress.name}, not plotting!")
                    continue

                stresses.loc[istress.name, :] = (xi, yi, name, f"C{count%10}")
            count += 1

        # create figure
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
        else:
            fig = ax.figure

        # add oseries
        osize = 50
        oserieslabel = ml.oseries.name

        if metadata_source == "model":
            xm = float(ml.oseries.metadata["x"])
            ym = float(ml.oseries.metadata["y"])
        elif metadata_source == "store":
            ometa = self.pstore.get_metadata("oseries", ml.oseries.name, as_frame=False)
            xm = float(ometa.pop("x", np.nan))
            ym = float(ometa.pop("y", np.nan))
        else:
            raise ValueError("metadata_source must be either " "'model' or 'store'!")

        po = ax.scatter(xm, ym, s=osize, marker="o", label=oserieslabel, color="k")
        legend_list = [po]

        # add stresses
        ax.scatter(
            stresses["x"],
            stresses["y"],
            s=50,
            c=stresses.color,
            marker="o",
            edgecolors="k",
            linewidths=0.75,
        )

        # label oseries
        if label:
            stroke = [patheffects.withStroke(linewidth=3, foreground="w")]
            txt = ax.annotate(
                text=oserieslabel,
                xy=(xm, ym),
                fontsize=8,
                textcoords="offset points",
                xytext=(10, 10),
            )
            txt.set_path_effects(stroke)

        # get legend entries for stressmodels
        uniques = stresses.loc[:, ["stressmodel", "color"]].drop_duplicates(
            keep="first"
        )
        for name, row in uniques.iterrows():
            (h,) = ax.plot(
                [],
                [],
                marker="o",
                label=row.stressmodel,
                ls="",
                mec="k",
                ms=10,
                color=row.color,
            )
            legend_list.append(h)

        # add legend
        ax.legend(legend_list, [i.get_label() for i in legend_list], loc="best")

        # set axes properties
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        for label in ax.get_yticklabels():
            label.set_rotation(90)
            label.set_verticalalignment("center")

        if offset > 0.0:
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            ax.set_xlim(xmin - offset, xmax + offset)
            ax.set_ylim(ymin - offset, ymax + offset)

        # label stresses
        if label:
            for name, row in stresses.iterrows():
                namestr = str(name)
                namestr += f"\n({row.stressmodel})"
                txt = ax.annotate(
                    text=namestr,
                    xy=(row.x, row.y),
                    fontsize=8,
                    textcoords="offset points",
                    xytext=(10, 10),
                )
                txt.set_path_effects(stroke)

        if backgroundmap:
            self.add_background_map(ax)

        fig.tight_layout()

        return ax

    def stresslinks(
        self,
        kinds=None,
        model_names=None,
        color_lines=False,
        alpha=0.4,
        ax=None,
        figsize=(10, 8),
        legend=True,
        labels=False,
        adjust=False,
        backgroundmap=False,
    ):
        """Create a map linking models with their stresses.

        Parameters
        ----------
        kinds: list, optional
            kinds of stresses to plot, defaults to None, which selects
            all kinds.
        model_names: list, optional
            list of model names to plot, substrings of model names
            are also accepted, defaults to None, which selects all
            models.
        color_lines: bool, optional
            if True, connecting lines have the same colors as the stresses,
            defaults to False, which uses a black line.
        alpha: float, optional
            alpha value for the connecting lines, defaults to 0.4.
        ax : matplotlib.Axes, optional
            axes handle, if not provided a new figure is created.
        figsize : tuple, optional
            figure size, by default (10, 8)
        legend: bool, optional
            create a legend for all unique kinds, defaults to True.
        labels: bool, optional
            add labels for stresses and oseries, defaults to False.
        adjust: bool, optional
            automated smart label placement using adjustText, by
            default False
        backgroundmap: bool, optional
            if True, add background map (default CRS is EPSG:28992) with default tiles
            by OpenStreetMap.Mapnik. Default option is False.

        Returns
        -------
        ax: axes object
            axis handle of the resulting figure

        See also
        --------
        self.add_background_map
        """
        if model_names:
            m_idx = self.pstore.search(libname="models", s=model_names)
        else:
            m_idx = self.pstore.model_names
        struct = self.pstore.get_model_timeseries_names(progressbar=False).loc[m_idx]

        oseries = self.pstore.oseries
        stresses = self.pstore.stresses
        skind = stresses.kind.unique()
        stused = np.array([])
        if kinds is None:
            kinds = skind

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        segments = []
        segment_colors = []
        scatter_colors = []
        ax.scatter(
            oseries.loc[struct["oseries"], "x"],
            oseries.loc[struct["oseries"], "y"],
            color="C0",
        )
        for m in struct.index:
            os = oseries.loc[struct.loc[m, "oseries"]]
            mstresses = struct.loc[m].drop("oseries").dropna().index
            st = stresses.loc[mstresses]
            for s in mstresses:
                if np.isin(st.loc[s, "kind"], kinds):
                    (c,) = np.where(skind == st.loc[s, "kind"])
                    if color_lines:
                        color = f"C{c[0]+1}"
                    else:
                        color = "k"
                    segments.append(
                        [[os["x"], os["y"]], [st.loc[s, "x"], st.loc[s, "y"]]]
                    )
                    segment_colors.append(color)
                    scatter_colors.append(f"C{c[0]+1}")

                    stused = np.append(stused, s)

        if labels:
            self.add_labels(oseries.loc[struct["oseries"].unique()], ax, adjust=adjust)
            self.add_labels(stresses.loc[np.unique(stused)], ax, adjust=adjust)

        ax.scatter(
            [x[1][0] for x in segments],
            [y[1][1] for y in segments],
            color=scatter_colors,
        )
        ax.add_collection(
            LineCollection(segments, colors=segment_colors, linewidths=0.5, alpha=alpha)
        )

        if legend:
            legend_elements = [
                Line2D(
                    [],
                    [],
                    marker="o",
                    color="w",
                    markerfacecolor="C0",
                    label="oseries",
                    markersize=10,
                )
            ]
            for kind in kinds:
                (c,) = np.where(skind == kind)
                legend_elements.append(
                    Line2D(
                        [],
                        [],
                        marker="o",
                        color="w",
                        markerfacecolor=f"C{c[0]+1}",
                        label=kind,
                        markersize=10,
                    )
                )
            ax.legend(handles=legend_elements)

        if backgroundmap:
            self.add_background_map(ax)

        return ax

    @staticmethod
    def _list_contextily_providers():
        """List contextily providers.

        Taken from contextily notebooks.

        Returns
        -------
        providers : dict
            dictionary containing all providers. See keys for names
            that can be passed as map_provider arguments.
        """
        import contextily as ctx

        providers = {}

        def get_providers(provider):
            if "url" in provider:
                providers[provider["name"]] = provider
            else:
                for prov in provider.values():
                    get_providers(prov)

        get_providers(ctx.providers)
        return providers

    @staticmethod
    def add_background_map(
        ax, proj="epsg:28992", map_provider="OpenStreetMap.Mapnik", **kwargs
    ):
        """Add background map to axes using contextily.

        Parameters
        ----------
        ax: matplotlib.Axes
            axes to add background map to
        map_provider: str, optional
            name of map provider, see `contextily.providers` for options.
            Default is 'OpenStreetMap.Mapnik'
        proj: pyproj.Proj or str, optional
            projection for background map, default is 'epsg:28992'
            (RD Amersfoort, a projection for the Netherlands)
        """
        import contextily as ctx

        if isinstance(proj, str):
            import pyproj

            proj = pyproj.Proj(proj)

        providers = Maps._list_contextily_providers()
        ctx.add_basemap(ax, source=providers[map_provider], crs=proj.srs, **kwargs)

    @staticmethod
    def add_labels(df, ax, adjust=False, **kwargs):
        """Add labels to points on plot.

        Uses dataframe index to label points.

        Parameters
        ----------
        df: pd.DataFrame
            DataFrame containing x, y - data. Index is used as label
        ax: matplotlib.Axes
            axes object to label points on
        adjust: bool
            automated smart label placement using adjustText
        **kwargs:
            keyword arguments to ax.annotate
        """
        stroke = [patheffects.withStroke(linewidth=3, foreground="w")]

        fontsize = kwargs.pop("fontsize", 10)

        if adjust:
            from adjustText import adjust_text

            texts = []
            for name, row in df.iterrows():
                texts.append(
                    ax.text(
                        row["x"],
                        row["y"],
                        name,
                        fontsize=fontsize,
                        **{"path_effects": stroke},
                    )
                )

            adjust_text(
                texts,
                force_text=0.05,
                **{
                    "arrowprops": {
                        "arrowstyle": "-",
                        "color": "k",
                        "alpha": 0.5,
                    }
                },
            )

        else:
            textcoords = kwargs.pop("textcoords", "offset points")
            xytext = kwargs.pop("xytext", (10, 10))

            for name, row in df.iterrows():
                namestr = str(name)
                ax.annotate(
                    text=namestr,
                    xy=(row["x"], row["y"]),
                    fontsize=fontsize,
                    textcoords=textcoords,
                    xytext=xytext,
                    **{"path_effects": stroke},
                )



import pastas as ps
from packaging.version import parse as parse_version

PASTAS_VERSION = parse_version(ps.__version__)
PASTAS_LEQ_022 = PASTAS_VERSION <= parse_version("0.22.0")

__version__ = "1.3.0"



import os
import tempfile
from contextlib import contextmanager

import pytest
from pytest_dependency import depends

import pastastore as pst


@contextmanager
def tempyaml(yaml):
    temp = tempfile.NamedTemporaryFile(delete=False)
    temp.write(yaml.encode("utf-8"))
    temp.close()
    try:
        yield temp.name
    finally:
        os.unlink(temp.name)


@pytest.mark.dependency()
def test_load_yaml_rechargemodel(pstore):
    yamlstr = """
    my_first_model:                   # model name
      oseries: oseries2               # head time series name, obtained from pastastore
      stressmodels:                   # stressmodels dictionary
        recharge:                     # name of the recharge stressmodel
          class: RechargeModel        # type of pastas StressModel
          prec: prec2                 # name of precipitation stress, obtained from pastastore
          evap: evap2                 # name of evaporation stress, obtained from pastastore
          recharge: Linear            # pastas recharge type
          rfunc: Exponential          # response function
    """
    with tempyaml(yamlstr) as f:
        ml = pstore.yaml.load(f)[0]
    pstore.add_model(ml)


@pytest.mark.dependency()
def test_load_yaml_stressmodel(pstore):
    yamlstr = """
    my_second_model:                  # model name
      oseries: oseries2               # head time series name, obtained from pastastore
      stressmodels:                   # stressmodels dictionary
        prec:                         # name of the recharge stressmodel
          class: StressModel          # type of pastas StressModel
          stress: prec2               # name of precipitation stress, obtained from pastastore
          rfunc: Gamma                # response function
    """
    with tempyaml(yamlstr) as f:
        ml = pstore.yaml.load(f)[0]
    pstore.add_model(ml)


@pytest.mark.dependency()
def test_load_yaml_wellmodel(pstore):
    yamlstr = """
    my_third_model:                   # model name
      oseries: oseries1               # head time series name, obtained from pastastore
      stressmodels:                   # stressmodels dictionary
        well:                         # name of the recharge stressmodel
          class: WellModel            # type of pastas StressModel
          stress: well1               # name of well stress, obtained from pastastore
          distances: [100]

    """
    with tempyaml(yamlstr) as f:
        ml = pstore.yaml.load(f)[0]
    pstore.add_model(ml)


@pytest.mark.dependency()
def test_write_load_compare_yaml(request, pstore):
    depends(
        request,
        [
            f"test_load_yaml_rechargemodel[{pstore.type}]",
            f"test_load_yaml_stressmodel[{pstore.type}]",
            f"test_load_yaml_wellmodel[{pstore.type}]",
        ],
    )
    pstore.yaml.export_models(modelnames=["my_first_model"])
    ml1 = pstore.models["my_first_model"]
    ml2 = pstore.yaml.load("my_first_model.yaml")[0]
    assert (
        pst.util.compare_models(ml1, ml2, detailed_comparison=True).iloc[1:, -1].all()
    )
    os.remove("my_first_model.yaml")


@pytest.mark.dependency()
def test_write_yaml_per_oseries(request, pstore):
    depends(
        request,
        [
            f"test_load_yaml_rechargemodel[{pstore.type}]",
            f"test_load_yaml_stressmodel[{pstore.type}]",
            f"test_load_yaml_wellmodel[{pstore.type}]",
        ],
    )
    pstore.yaml.export_stored_models_per_oseries()
    os.remove("oseries1.yaml")
    os.remove("oseries2.yaml")


@pytest.mark.dependency()
def test_write_yaml_minimal(request, pstore):
    depends(
        request,
        [
            f"test_load_yaml_rechargemodel[{pstore.type}]",
            f"test_load_yaml_stressmodel[{pstore.type}]",
            f"test_load_yaml_wellmodel[{pstore.type}]",
        ],
    )
    ml = pstore.models["my_first_model"]
    pstore.yaml.export_model(ml, minimal_yaml=True)
    os.remove("my_first_model.yaml")


@pytest.mark.dependency()
def test_write_yaml_minimal_nearest(request, pstore):
    depends(
        request,
        [
            f"test_load_yaml_rechargemodel[{pstore.type}]",
            f"test_load_yaml_stressmodel[{pstore.type}]",
            f"test_load_yaml_wellmodel[{pstore.type}]",
        ],
    )
    ml = pstore.models["my_third_model"]
    pstore.yaml.export_model(ml, minimal_yaml=True, use_nearest=True)
    os.remove("my_third_model.yaml")



import warnings

import numpy as np
import pandas as pd
import pastas as ps
import pytest
from pytest_dependency import depends

with warnings.catch_warnings():
    warnings.simplefilter(action="ignore", category=FutureWarning)
    import pastastore as pst

ps.set_log_level("ERROR")


def test_get_library(conn):
    _ = conn._get_library("oseries")


def test_add_get_series(request, conn):
    o1 = pd.Series(
        index=pd.date_range("2000", periods=10, freq="D"),
        data=1.0,
        dtype=np.float64,
    )
    o1.name = "test_series"
    conn.add_oseries(o1, "test_series", metadata=None)
    o2 = conn.get_oseries("test_series")
    # PasConnector has no logic for preserving Series
    if conn.conn_type == "pas":
        o2 = o2.squeeze()
    try:
        assert isinstance(o2, pd.Series)
        assert (o1 == o2).all()
    finally:
        conn.del_oseries("test_series")


def test_add_get_series_wnans(request, conn):
    o1 = pd.Series(
        index=pd.date_range("2000", periods=10, freq="D"),
        data=1.0,
        dtype=np.float64,
    )
    o1.iloc[-3:] = np.nan
    o1.name = "test_series_nans"
    conn.add_oseries(o1, "test_series_nans", metadata=None)
    o2 = conn.get_oseries("test_series_nans")
    # PasConnector has no logic for preserving Series
    if conn.conn_type == "pas":
        o2 = o2.squeeze()
    try:
        assert isinstance(o2, pd.Series)
        assert o1.equals(o2)
    finally:
        conn.del_oseries("test_series_nans")


def test_add_get_dataframe(request, conn):
    o1 = pd.DataFrame(
        data=1.0,
        columns=["test_df"],
        index=pd.date_range("2000", periods=10, freq="D"),
    )
    o1.index.name = "test_idx"
    conn.add_oseries(o1, "test_df", metadata=None)
    o2 = conn.get_oseries("test_df")
    try:
        assert isinstance(o2, pd.DataFrame)
        # little hack as PasConnector has dtype int after load...
        assert o1.equals(o2.astype(float))
    finally:
        conn.del_oseries("test_df")


def test_add_pastas_timeseries(request, conn):
    o1 = pd.DataFrame(
        data=1.0,
        columns=["test_df"],
        index=pd.date_range("2000", periods=10, freq="D"),
    )
    o1.index.name = "test_idx"
    ts = ps.timeseries.TimeSeries(o1, metadata={"x": 100000.0, "y": 400000.0})
    try:
        conn.add_oseries(ts, "test_pastas_ts", metadata=None)
    except DeprecationWarning:
        pass


def test_update_series(request, conn):
    o1 = pd.DataFrame(
        data=1.0,
        columns=["test_df"],
        index=pd.date_range("2000", periods=10, freq="D"),
    )
    o1.index.name = "test_idx"
    conn.add_oseries(o1, "test_df", metadata={"x": 100000.0})
    o2 = pd.DataFrame(
        data=2.0,
        columns=["test_df"],
        index=pd.date_range("2000-01-10", periods=2, freq="D"),
    )
    o2.index.name = "test_idx"
    conn.update_oseries(o2, "test_df", metadata={"x": 200000.0, "y": 400000})
    o3 = conn.get_oseries("test_df")
    try:
        assert (o3.iloc[-2:] == 2.0).all().all()
        assert o3.index.size == 11
    finally:
        conn.del_oseries("test_df")


def test_upsert_oseries(request, conn):
    o1 = pd.DataFrame(
        data=1.0,
        columns=["test_df"],
        index=pd.date_range("2000", periods=10, freq="D"),
    )
    o1.index.name = "test_idx"
    conn.upsert_oseries(o1, "test_df", metadata={"x": 100000.0})
    o2 = pd.DataFrame(
        data=2.0,
        columns=["test_df"],
        index=pd.date_range("2000-01-05", periods=10, freq="D"),
    )
    o2.index.name = "test_idx"
    conn.upsert_oseries(o2, "test_df", metadata={"x": 200000.0, "y": 400000})
    o3 = conn.get_oseries("test_df")
    try:
        assert (o3.iloc[-10:] == 2.0).all().all()
        assert o3.index.size == 14
    finally:
        conn.del_oseries("test_df")


def test_upsert_stress(request, conn):
    s1 = pd.DataFrame(
        data=1.0,
        columns=["test_df"],
        index=pd.date_range("2000", periods=10, freq="D"),
    )
    s1.index.name = "test_idx"
    conn.upsert_stress(s1, "test_df", kind="useless", metadata={"x": 100000.0})
    s2 = pd.DataFrame(
        data=2.0,
        columns=["test_df"],
        index=pd.date_range("2000-01-05", periods=10, freq="D"),
    )
    s2.index.name = "test_idx"
    conn.upsert_stress(
        s2,
        "test_df",
        kind="not useless",
        metadata={"x": 200000.0, "y": 400000},
    )
    s3 = conn.get_stresses("test_df")
    try:
        assert (s3.iloc[-10:] == 2.0).all().all()
        assert s3.index.size == 14
        assert conn.stresses.loc["test_df", "kind"] == "not useless"
    finally:
        conn.del_stress("test_df")


def test_update_metadata(request, conn):
    o1 = pd.DataFrame(
        data=1.1,
        columns=["test_df"],
        index=pd.date_range("2000", periods=10, freq="D"),
        dtype=float,
    )
    o1.index.name = "test_idx"
    conn.add_oseries(o1, "test_df", metadata={"x": 100000.0})
    conn.update_metadata("oseries", "test_df", {"x": 200000.0, "y": 400000.0})
    m = conn._get_metadata("oseries", "test_df")
    try:
        assert isinstance(m, dict)
        assert m["x"] == 200000.0
        assert m["y"] == 400000.0
    finally:
        conn.del_oseries("test_df")


@pytest.mark.dependency()
def test_add_oseries(conn):
    o = pd.read_csv("./tests/data/obs.csv", index_col=0, parse_dates=True)
    conn.add_oseries(
        o,
        "oseries1",
        metadata={"name": "oseries1", "x": 100000, "y": 400000},
        overwrite=True,
    )


@pytest.mark.dependency()
def test_add_stress(conn):
    s = pd.read_csv("./tests/data/rain.csv", index_col=0, parse_dates=True)
    conn.add_stress(
        s,
        "prec",
        kind="prec",
        metadata={"kind": "prec", "x": 100001, "y": 400001},
    )


@pytest.mark.dependency()
def test_get_oseries(request, conn):
    depends(request, [f"test_add_oseries[{conn.type}]"])
    _ = conn.get_oseries("oseries1")


@pytest.mark.dependency()
def test_get_oseries_and_metadata(request, conn):
    depends(request, [f"test_add_oseries[{conn.type}]"])
    _ = conn.get_oseries("oseries1", return_metadata=True)


@pytest.mark.dependency()
def test_get_stress(request, conn):
    depends(request, [f"test_add_stress[{conn.type}]"])
    s = conn.get_stresses("prec")
    s.name = "prec"


@pytest.mark.dependency()
def test_get_stress_and_metadata(request, conn):
    depends(request, [f"test_add_stress[{conn.type}]"])
    s, _ = conn.get_stresses("prec", return_metadata=True)
    s.name = "prec"


@pytest.mark.dependency()
def test_oseries_prop(request, conn):
    depends(request, [f"test_add_oseries[{conn.type}]"])
    conn.oseries


@pytest.mark.dependency()
def test_stresses_prop(request, conn):
    depends(request, [f"test_add_stress[{conn.type}]"])
    conn.stresses


def test_repr(conn):
    conn.__repr__()


@pytest.mark.dependency()
def test_del_oseries(request, conn):
    depends(request, [f"test_add_oseries[{conn.type}]"])
    conn.del_oseries("oseries1")


@pytest.mark.dependency()
def test_del_stress(request, conn):
    depends(request, [f"test_add_stress[{conn.type}]"])
    conn.del_stress("prec")


@pytest.mark.dependency()
def test_empty_library(request, conn):
    s1 = pd.Series(
        index=pd.date_range("2000", periods=10, freq="D"),
        data=1.0,
        dtype=np.float64,
    )
    s1.name = "test_series"
    conn.add_oseries(s1, "test_series", metadata=None)
    conn.empty_library("stresses", prompt=False, progressbar=False)


@pytest.mark.dependency()
def test_delete(request, conn):
    # no need to delete dictconnector (in memory)
    if conn.conn_type == "arctic":
        pst.util.delete_arctic_connector(conn, libraries=["oseries"])
        pst.util.delete_arctic_connector(conn)
    elif conn.conn_type == "pystore":
        pst.util.delete_pystore_connector(conn, libraries=["oseries"])
        pst.util.delete_pystore_connector(conn)
    elif conn.conn_type == "pas":
        pst.util.delete_pas_connector(conn, libraries=["oseries"])
        pst.util.delete_pas_connector(conn)



import datetime
import logging
import os
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import pastas as ps
import yaml

from pastastore.version import PASTAS_LEQ_022

ps.logger.setLevel("ERROR")

logging.basicConfig(level="INFO")
logger = logging.getLogger(__name__)


def _convert_dict_dtypes_for_yaml(d: Dict[str, Any]):
    """Internal method to convert dictionary values for storing in YAML format.

    Parameters
    ----------
    d : dict
        dictionary to parse iteratively
    """
    for k, v in d.items():
        if isinstance(v, dict):
            _convert_dict_dtypes_for_yaml(v)
        elif isinstance(v, list) and k == "stress":
            for iv in v:
                if isinstance(iv, dict):
                    _convert_dict_dtypes_for_yaml(iv)
        elif isinstance(v, pd.Timestamp):
            d[k] = v.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(v, datetime.datetime):
            d[k] = pd.to_datetime(v).strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(v, pd.Timedelta):
            d[k] = v.to_timedelta64().__str__()
        elif isinstance(v, datetime.timedelta):
            d[k] = pd.to_timedelta(v).to_timedelta64().__str__()
        elif isinstance(v, np.int64):
            d[k] = int(v)
        elif isinstance(v, np.float64):
            d[k] = float(v)
        elif isinstance(v, pd.DataFrame):
            d[k] = v.reset_index().to_dict(orient="records")


def replace_ts_with_name(d, nearest=False):
    """Replace time series dict with its name in pastas model dict.

    Parameters
    ----------
    d : dict
        pastas model dictionary
    nearest : bool, optional
        replace time series with "nearest" option. Warning, this does not
        check whether the time series are actually the nearest ones!
    """
    for k, v in d.items():
        if k in ["oseries", "prec", "evap", "stress"]:
            if isinstance(v, dict):
                if nearest and k != "oseries":
                    if k == "stress":
                        d[k] = "nearest <kind>"
                    else:
                        d[k] = f"nearest {k}"
                else:
                    d[k] = v["name"]
            elif isinstance(v, list):
                if nearest:
                    d[k] = f"nearest {len(v)} well"
                else:
                    d[k] = [iv["name"] for iv in v]
        elif isinstance(v, dict):
            replace_ts_with_name(v, nearest=nearest)


def reduce_to_minimal_dict(d, keys=None):
    """Reduce pastas model dictionary to a minimal form.

    This minimal form strives to keep the minimal information that still
    allows a model to be constructed. Users are warned, reducing a model
    dictionary with this function can lead to a different model than
    the original!

    Parameters
    ----------
    d : dict
        pastas model in dictionary form
    keys : list, optional
        list of keys to keep, by default None, which defaults to:
        ["name", "oseries", "settings", "tmin", "tmax", "noise",
        "stressmodels", "rfunc", "stress", "prec", "evap", "stressmodel"]
    """

    if keys is None:
        keys = [
            "name",
            "oseries",
            "settings",
            "tmin",
            "tmax",
            "noise",
            "stressmodels",
            "rfunc",
            "stress",
            "prec",
            "evap",
            "stressmodel" if PASTAS_LEQ_022 else "class",
        ]

    # also keep stressmodels by adding names to keys list
    if "stressmodels" in d:
        keys += list(d["stressmodels"].keys())

    iter_d = d.copy()  # copy dictionary to use as iterator

    # delete keys if not in keys list or if value is None
    for k, v in iter_d.items():
        if (k not in keys) or (v is None):
            del d[k]
        elif isinstance(v, dict):
            reduce_to_minimal_dict(v, keys=keys)


class PastastoreYAML:
    """Class for reading/writing Pastas models in YAML format.

    This class provides a more human-readable form of Pastas models in
    comparison to Pastas default .pas (JSON) files. The goal is to provide
    users with a simple mini-language to quickly build/test different model
    structures. A PastaStore is required as input, which contains existing
    models or time series required to build new models. This
    class also introduces some shortcuts to simplify building models.
    Shortcuts include the option to pass 'nearest' as the name of a stress,
    which will automatically select the closest stress of a particular type.
    Other shortcuts include certain default options when certain information
    is not listed in the YAML file, that will work well in many cases.

    Usage
    -----
    Instantiate the PastastoreYAML class::

        pyaml = PastastoreYAML(pstore)

    Export a Pastas model to a YAML file::

        pyaml.export_model_to_yaml(ml)

    Load a Pastas model from a YAML file::

        models = pyaml.load_yaml("my_first_model.yaml")

    Example YAML file using 'nearest'::

      my_first_model:  # this is the name of the model
        oseries: "oseries1"  # name of oseries stored in PastaStore
        stressmodels:
            recharge:  # recognized as RechargeModel by name
              prec: "nearest"  # use nearest stress with kind="prec"
              evap: "EV24_DEELEN"  # specific station
            river:
              stress: "nearest riv"  # nearest stress with kind="riv"
            wells:
              stress: "nearest 3"  # nearest 3 stresses with kind="well"
              stressmodel: WellModel  # provide StressModel type
    """

    def __init__(self, pstore):
        """Constructor for PastasstoreYAML class.

        Parameters
        ----------
        pstore : pastastore.PastaStore
            PastaStore object containing models to be exported as YAML files
            or containing time series that are referenced in YAML files.
        """
        self.pstore = pstore

    def _parse_rechargemodel_dict(self, d: Dict, onam: Optional[str] = None) -> Dict:
        """Internal method to parse RechargeModel dictionary.

        Note: supports 'nearest' as input to 'prec' and 'evap',
        which will automatically select nearest stress with kind="prec" or
        "evap". Requires "x" and "y" locations to be present in both oseries
        and stresses metadata.

        Parameters
        ----------
        d : dict
            dictionary containing RechargeModel information
        onam : str, optional
            name of oseries used when 'nearest' is provided as prec or evap,
            by default None

        Returns
        -------
        d : dict
            dictionary that can be read by ps.io.base.load(),
            containing stresses obtained from PastaStore, and setting
            defaults if they were not already provided.
        """
        # precipitation
        prec_val = d.get("prec", "nearest")
        if isinstance(prec_val, dict):
            pnam = prec_val["name"]
            p = self.pstore.get_stresses(pnam)
            prec_val["series"] = p
            prec = prec_val
        elif prec_val.startswith("nearest"):
            if onam is None:
                raise ValueError("Provide oseries name when using nearest!")
            if len(prec_val.split()) > 1:
                kind = prec_val.split()[-1]
            else:
                kind = "prec"
            pnam = self.pstore.get_nearest_stresses(onam, kind=kind).iloc[0, 0]
            logger.info(f"  | using nearest stress with kind='{kind}': '{pnam}'")
            p, pmeta = self.pstore.get_stresses(pnam, return_metadata=True)
            prec = {
                "name": pnam,
                "settings": "prec",
                "metadata": pmeta,
                "series": p,
            }
        elif isinstance(prec_val, str):
            pnam = d["prec"]
            p, pmeta = self.pstore.get_stresses(pnam, return_metadata=True)
            prec = {
                "name": pnam,
                "settings": "prec",
                "metadata": pmeta,
                "series": p,
            }
        else:
            raise NotImplementedError(f"Could not parse prec value: '{prec_val}'")
        d["prec"] = prec

        # evaporation
        evap_val = d.get("evap", "nearest")
        if isinstance(evap_val, dict):
            enam = evap_val["name"]
            e = self.pstore.get_stresses(enam)
            evap_val["series"] = e
            evap = evap_val
        elif evap_val.startswith("nearest"):
            if onam is None:
                raise ValueError("Provide oseries name when using nearest!")
            if len(evap_val.split()) > 1:
                kind = evap_val.split()[-1]
            else:
                kind = "evap"
            enam = self.pstore.get_nearest_stresses(onam, kind=kind).iloc[0, 0]
            logger.info(f"  | using nearest stress with kind='{kind}': '{enam}'")
            e, emeta = self.pstore.get_stresses(enam, return_metadata=True)
            evap = {
                "name": enam,
                "settings": "evap",
                "metadata": emeta,
                "series": e,
            }
        elif isinstance(evap_val, str):
            enam = d["evap"]
            e, emeta = self.pstore.get_stresses(enam, return_metadata=True)
            evap = {
                "name": enam,
                "settings": "evap",
                "metadata": emeta,
                "series": e,
            }
        else:
            raise NotImplementedError(f"Could not parse evap value: '{evap_val}'")
        d["evap"] = evap

        # rfunc
        if "rfunc" not in d:
            logger.info("  | no 'rfunc' provided, using 'Exponential'")
        # for pastas >= 0.23.0, convert rfunc value to dictionary with 'class' key
        elif not isinstance(d["rfunc"], dict) and not PASTAS_LEQ_022:
            d["rfunc"] = {"class": d["rfunc"]}

        # stressmodel
        classkey = "stressmodel" if PASTAS_LEQ_022 else "class"
        if classkey not in d:
            d[classkey] = "RechargeModel"

        # recharge type (i.e. Linear, FlexModel, etc.)
        if ("recharge" not in d) and (d[classkey] == "RechargeModel"):
            logger.info("  | no 'recharge' type provided, using 'Linear'")
        # if pastas >= 0.23.0, recharge value must be dict with class key
        elif not isinstance(d["recharge"], dict) and not PASTAS_LEQ_022:
            d["recharge"] = {"class": d["recharge"]}

        # tarsomodel logic
        if d[classkey] == "TarsoModel":
            dmin = d.get("dmin", None)
            dmax = d.get("dmin", None)
            oseries = d.get("oseries", None)
            if ((dmin is None) or (dmax is None)) and (oseries is None):
                logger.info(
                    "  | no 'dmin/dmax' or 'oseries' provided,"
                    f" filling in 'oseries': '{onam}'"
                )
                d["oseries"] = onam

        if "oseries" in d:
            onam = d["oseries"]
            if isinstance(onam, str):
                o = self.pstore.get_oseries(onam)
                d["oseries"] = o

        return d

    def _parse_stressmodel_dict(self, d: Dict, onam: Optional[str] = None) -> Dict:
        """Internal method to parse StressModel dictionary.

        Note: supports 'nearest' or 'nearest <kind>' as input to 'stress',
        which will automatically select nearest stress with kind=<kind>.
        Requires "x" and "y" locations to be present in both oseries and
        stresses metadata.

        Parameters
        ----------
        d : dict
            dictionary containing WellModel information
        onam : str, optional
            name of oseries used when 'nearest <kind>' is provided as stress,
            by default None

        Returns
        -------
        d : dict
            dictionary that can be read by ps.io.base.load(),
            containing stresses obtained from PastaStore, and setting
            defaults if they were not already provided.
        """

        # get stress
        snam = d.pop("stress")

        # if list, obtain first and only entry
        if isinstance(snam, list):
            snam = snam[0]
        # if str, either name of single series or 'nearest <kind>'
        if snam.startswith("nearest"):
            if len(snam.split()) > 1:
                kind = snam.split()[-1]
            else:
                kind = None
            if kind == "oseries":
                snam = self.pstore.get_nearest_oseries(onam).iloc[0, 0]
            else:
                snam = self.pstore.get_nearest_stresses(onam, kind=kind).iloc[0, 0]
            logger.info(f"  | using nearest stress with kind='{kind}': {snam}")

        s, smeta = self.pstore.get_stresses(snam, return_metadata=True)
        s = {
            "name": snam,
            "settings": d.pop("settings", None),
            "metadata": smeta,
            "series": s,
        }
        d["stress"] = [s] if PASTAS_LEQ_022 else s

        # use stress name if not provided
        if "name" not in d:
            d["name"] = snam

        # rfunc
        if "rfunc" not in d:
            logger.info("  | no 'rfunc' provided, using 'Gamma'")
            d["rfunc"] = "Gamma" if PASTAS_LEQ_022 else {"class": "Gamma"}
        # for pastas >= 0.23.0, convert rfunc value to dictionary with 'class' key
        elif not isinstance(d["rfunc"], dict) and not PASTAS_LEQ_022:
            d["rfunc"] = {"class": d["rfunc"]}

        return d

    def _parse_wellmodel_dict(self, d: Dict, onam: Optional[str] = None) -> Dict:
        """Internal method to parse WellModel dictionary.

        Note: supports 'nearest' or 'nearest <number> <kind>' as input to
        'stress', which will automatically select nearest or <number> of
        nearest stress(es) with kind=<kind>. Requires "x" and "y" locations to
        be present in both oseries and stresses metadata.

        Parameters
        ----------
        d : dict
            dictionary containing WellModel information
        onam : str, optional
            name of oseries used when 'nearest' is provided as stress,
            by default None

        Returns
        -------
        d : dict
            dictionary that can be read by ps.io.base.load(),
            containing stresses obtained from PastaStore, and setting
            defaults if they were not already provided.
        """

        # parse stress
        snames = d.pop("stress")

        # if str, either name of single series or 'nearest <n> <kind>'
        if isinstance(snames, str):
            if snames.startswith("nearest"):
                if len(snames.split()) == 3:
                    n = int(snames.split()[1])
                    kind = snames.split()[2]
                elif len(snames.split()) == 2:
                    try:
                        n = int(snames.split()[1])
                    except ValueError:
                        raise ValueError(
                            f"Could not parse: '{snames}'! "
                            "When using option 'nearest' for WellModel,  "
                            "use 'nearest <n>' or 'nearest <n> <kind>'!"
                        )
                    kind = "well"
                elif len(snames.split()) == 1:
                    n = 1
                    kind = "well"
                snames = (
                    self.pstore.get_nearest_stresses(onam, kind=kind, n=n)
                    .iloc[0]
                    .values
                )
                logger.info(
                    f"  | using {n} nearest stress(es) with kind='{kind}': " f"{snames}"
                )
            else:
                snames = [snames]

        # get time series
        slist = []
        for snam in snames:
            s, smeta = self.pstore.get_stresses(snam, return_metadata=True)
            sdict = {
                "name": snam,
                "settings": "well",
                "metadata": smeta,
                "series": s,
            }
            slist.append(sdict)
        d["stress"] = slist

        # get distances
        if "distances" not in d:
            d["distances"] = self.pstore.get_distances(
                oseries=onam, stresses=snames
            ).values.squeeze()

        # use default name if not provided
        if "name" not in d:
            d["name"] = "wells"

        # rfunc
        if "rfunc" not in d:
            logger.info("  | no 'rfunc' provided, using 'HantushWellModel'")
            d["rfunc"] = (
                "HantushWellModel" if PASTAS_LEQ_022 else {"class": "HantushWellModel"}
            )
        # for pastas >= 0.23.0, convert rfunc value to dictionary with 'class' key
        elif not isinstance(d["rfunc"], dict) and not PASTAS_LEQ_022:
            d["rfunc"] = {"class": d["rfunc"]}

        if "up" not in d:
            logger.info(
                "  | no 'up' provided, set to 'False', "
                "(i.e. pumping rate is positive for extraction)."
            )
            d["up"] = False

        return d

    def construct_mldict(self, mlyml: dict, mlnam: str) -> dict:
        # get oseries + metadata
        if isinstance(mlyml["oseries"], dict):
            onam = str(mlyml["oseries"]["name"])
            _ = mlyml.pop("oseries")
        else:
            onam = str(mlyml.pop("oseries"))

        logger.info(f"Building model '{mlnam}' for oseries '{onam}'")
        o, ometa = self.pstore.get_oseries(onam, return_metadata=True)

        # create model to obtain default model settings
        ml = ps.Model(o, name=mlnam, metadata=ometa)
        mldict = ml.to_dict(series=True)

        # update with stored model settings
        if "settings" in mlyml:
            mldict["settings"].update(mlyml["settings"])

        # stressmodels
        for smnam, smyml in mlyml["stressmodels"].items():
            # set name if not provided
            if smyml is not None:
                name = smyml.get("name", smnam)
            else:
                name = smnam
            logger.info(f"| parsing stressmodel: '{name}'")

            # check whether smtyp is defined
            classkey = "stressmodel" if PASTAS_LEQ_022 else "class"
            if smyml is not None:
                if PASTAS_LEQ_022:
                    if "class" in smyml:
                        smyml["stressmodel"] = smyml.pop("class")
                if classkey in smyml:
                    smtyp = True
                else:
                    smtyp = False
            else:
                smtyp = False

            # check if RechargeModel based on name if smtyp not defined
            if (
                smnam.lower() in ["rch", "rech", "recharge", "rechargemodel"]
            ) and not smtyp:
                logger.info("| assuming RechargeModel based on stressmodel name.")
                # check if stressmodel dictionary is empty, create (nearly
                # empty) dict so defaults are used
                if smyml is None:
                    mlyml["stressmodels"][smnam] = {"name": "recharge"}
                    smyml = mlyml["stressmodels"][smnam]
                if "name" not in smyml:
                    smyml["name"] = smnam
                smtyp = smyml.get(classkey, "RechargeModel")
            else:
                # if no info is provided, raise error,
                # cannot make any assumptions for non-RechargeModels
                if smyml is None:
                    raise ValueError(
                        "Insufficient information " f"for stressmodel '{name}'!"
                    )
                # get stressmodel type, with default StressModel
                if classkey in smyml:
                    smtyp = smyml[classkey]
                else:
                    logger.info(
                        "| no stressmodel class type provided, " "using 'StressModel'"
                    )
                    smtyp = "StressModel"

            # parse dictionary based on smtyp
            if smtyp in ["RechargeModel", "TarsoModel"]:
                # parse RechargeModel
                sm = self._parse_rechargemodel_dict(smyml, onam=onam)

                # turn off constant for TarsoModel
                if smtyp == "TarsoModel":
                    mldict["constant"] = False
            elif smtyp == "StressModel":
                # parse StressModel
                sm = self._parse_stressmodel_dict(smyml, onam=onam)
            elif smtyp == "WellModel":
                # parse WellModel
                sm = self._parse_wellmodel_dict(smyml, onam=onam)
            else:
                raise NotImplementedError(
                    "PastaStore.yaml interface does " f"not (yet) support '{smtyp}'!"
                )

            # add to list
            smyml.update(sm)

        # update model dict w/ default settings with loaded data
        mldict.update(mlyml)

        # add name to dictionary if not already present
        if "name" not in mldict:
            mldict["name"] = mlnam

        # convert warmup and time_offset to panads.Timedelta
        if "warmup" in mldict["settings"]:
            mldict["settings"]["warmup"] = pd.Timedelta(mldict["settings"]["warmup"])
        if "time_offset" in mldict["settings"]:
            mldict["settings"]["time_offset"] = pd.Timedelta(
                mldict["settings"]["time_offset"]
            )
        return mldict

    def load(self, fyaml: str) -> List[ps.Model]:
        """Load Pastas YAML file.

        Note: currently supports RechargeModel, StressModel and WellModel.

        Parameters
        ----------
        fyaml : str
            path to file

        Returns
        -------
        models : list
            list containing pastas model(s)

        Raises
        ------
        ValueError
            if insufficient information is provided to construct pastas model
        NotImplementedError
            if unsupported stressmodel is encountered
        """

        with open(fyaml, "r") as f:
            yml = yaml.load(f, Loader=yaml.CFullLoader)

        models = []

        for mlnam in yml.keys():
            mlyml = yml[mlnam]

            mldict = self.construct_mldict(mlyml, mlnam)

            # load model
            ml = ps.io.base._load_model(mldict)
            models.append(ml)

        return models

    def export_stored_models_per_oseries(
        self,
        oseries: Optional[Union[List[str], str]] = None,
        outdir: Optional[str] = ".",
        minimal_yaml: Optional[bool] = False,
        use_nearest: Optional[bool] = False,
    ):
        """Export store models grouped per oseries (location) to YAML file(s).

        Note: The oseries names are used as file names.

        Parameters
        ----------
        oseries : list of str, optional
            list of oseries (location) names, by default None, which uses
            all stored oseries for which there are models.
        outdir : str, optional
            path to output directory, by default "." (current directory)
        minimal_yaml : bool, optional
            reduce yaml file to include the minimum amount of information
            that will still construct a model. Users are warned, using this
            option does not guarantee the same model will be constructed
            as the one that was exported! Default is False.
        use_nearest : bool, optional
            if True, replaces time series with "nearest <kind>", filling in
            kind where possible. Warning! This does not check whether
            the time series are actually the nearest ones! Only used
            when minimal_yaml=True. Default is False.
        """

        onames = self.pstore.conn._parse_names(oseries, "oseries")

        for onam in onames:
            try:
                onam = int(onam)
            except ValueError:
                pass
            # check if any models exist for oseries
            if onam not in self.pstore.oseries_models:
                continue

            model_list = self.pstore.get_models(
                self.pstore.oseries_models[onam],
                return_dict=True,
                squeeze=False,
            )

            model_dicts = {}
            for d in model_list:
                if minimal_yaml:
                    replace_ts_with_name(d, nearest=use_nearest)
                    reduce_to_minimal_dict(d)
                _convert_dict_dtypes_for_yaml(d)
                name = d.pop("name")
                model_dicts[name] = d

            with open(os.path.join(outdir, f"{onam}.yaml"), "w") as f:
                yaml.dump(model_dicts, f, Dumper=yaml.CDumper)

    def export_models(
        self,
        models: Optional[Union[List[ps.Model], List[Dict]]] = None,
        modelnames: Optional[Union[List[str], str]] = None,
        outdir: Optional[str] = ".",
        minimal_yaml: Optional[bool] = False,
        use_nearest: Optional[bool] = False,
        split: Optional[bool] = True,
        filename: Optional[str] = "pastas_models.yaml",
    ):
        """Export (stored) models to yaml file(s).

        Parameters
        ----------
        models : list of ps.Model or dict, optional
            pastas Models to write to yaml file(s), if not provided,
            uses modelnames to collect stored models to export.
        modelnames : list of str, optional
            list of model names to export, by default None, which uses
            all stored models.
        outdir : str, optional
            path to output directory, by default "." (current directory)
        minimal_yaml : bool, optional
            reduce yaml file to include the minimum amount of information
            that will still construct a model. Users are warned, using this
            option does not guarantee the same model will be constructed
            as the one that was exported! Default is False.
        use_nearest : bool, optional
            if True, replaces time series with "nearest <kind>", filling in
            kind where possible. Warning! This does not check whether
            the time series are actually the nearest ones! Only used
            when minimal_yaml=True. Default is False.
        split : bool, optional
            if True, split into separate yaml files, otherwise store all
            in the same file. The model names are used as file names.
        filename : str, optional
            filename for YAML file, only used if `split=False`
        """
        if models is None:
            modelnames = self.pstore.conn._parse_names(modelnames, "models")
            model_list = self.pstore.get_models(
                modelnames, return_dict=True, squeeze=False
            )
        else:
            model_list = [
                iml.to_dict(series=False) if isinstance(iml, ps.Model) else iml
                for iml in models
            ]

        # each model in separate file
        if split:
            for ml in model_list:
                self.export_model(ml, outdir=outdir, minimal_yaml=minimal_yaml)
        # all models in same file
        else:
            model_dicts = {}
            for d in model_list:
                if minimal_yaml:
                    replace_ts_with_name(d, nearest=use_nearest)
                    reduce_to_minimal_dict(d)
                _convert_dict_dtypes_for_yaml(d)
                name = d.pop("name")
                model_dicts[name] = d

            with open(os.path.join(outdir, filename), "w") as f:
                yaml.dump(model_dicts, f, Dumper=yaml.CDumper)

    @staticmethod
    def export_model(
        ml: Union[ps.Model, dict],
        outdir: Optional[str] = ".",
        minimal_yaml: Optional[bool] = False,
        use_nearest: Optional[bool] = False,
    ):
        """Write single pastas model to YAML file.

        Parameters
        ----------
        ml : ps.Model or dict
            pastas model instance or dictionary representing a pastas model
        outdir : str, optional
            path to output directory, by default "." (current directory)
        minimal_yaml : bool, optional
            reduce yaml file to include the minimum amount of information
            that will still construct a model. Users are warned, using this
            option does not guarantee the same model will be constructed
            as the one that was exported! Default is False.
        use_nearest : bool, optional
            if True, replaces time series with "nearest <kind>", filling in
            kind where possible. Warning! This does not check whether
            the time series are actually the nearest ones! Only used
            when minimal_yaml=True. Default is False.
        """
        if isinstance(ml, dict):
            name = ml["name"]
        else:
            name = ml.name
        with open(os.path.join(outdir, f"{name}.yaml"), "w") as f:
            if isinstance(ml, ps.Model):
                mldict = deepcopy(ml.to_dict(series=False))
            elif isinstance(ml, dict):
                mldict = ml
            else:
                raise TypeError("Only accepts dictionary or pastas.Model!")
            mlname = mldict.pop("name")
            if minimal_yaml:
                replace_ts_with_name(mldict, nearest=use_nearest)
                reduce_to_minimal_dict(mldict)
            _convert_dict_dtypes_for_yaml(mldict)
            yaml.dump({mlname: mldict}, f, Dumper=yaml.CDumper)



# ruff: noqa: F401
from pastastore import connectors, util
from pastastore.connectors import (
    ArcticConnector,
    ArcticDBConnector,
    DictConnector,
    PasConnector,
    PystoreConnector,
)
from pastastore.store import PastaStore
from pastastore.version import __version__
